use super::{VerbContext, UserVerb, UserVerbRef, UResult, UserError,
            get_player_item_or_fail, user_error, search_item_for_user};
use async_trait::async_trait;
use crate::{
    models::{
        item::{Item, DeathData, SkillType},
    },
    db::ItemSearchParams,
    static_content::possession_type::possession_data,
    language::join_words,
    services::{
        destroy_container,
        skills::skill_check_and_grind, comms::broadcast_to_room,
        capacity::{CapacityLevel, check_item_capacity}},
};
use ansi::ansi;

pub struct Verb;
#[async_trait]
impl UserVerb for Verb {
    async fn handle(self: &Self, ctx: &mut VerbContext, _verb: &str, remaining: &str) -> UResult<()> {
        let (what_raw, corpse_raw) = match remaining.split_once(" from ") {
            None => user_error(ansi!("Usage: <bold>cut<reset> thing <bold>from<reset> corpse").to_owned())?,
            Some(v) => v
        };

        let player_item = get_player_item_or_fail(ctx).await?;
        
        let corpse = search_item_for_user(ctx, &ItemSearchParams {
            include_loc_contents: true,
            dead_first: true,
            ..ItemSearchParams::base(&player_item, corpse_raw.trim())
        }).await?;

        let what_norm = what_raw.trim().to_lowercase();
        let possession_type = match corpse.death_data.as_ref() {
            None => user_error(format!("You can't do that while {} is still alive!", corpse.pronouns.subject))?,
            Some(DeathData { parts_remaining, ..}) =>
                parts_remaining.iter().find(
                    |pt| possession_data().get(pt)
                        .map(|pd| pd.display.to_lowercase() == what_norm ||
                                  pd.aliases.iter().any(|a| a.to_lowercase() == what_norm))
                        == Some(true)).ok_or_else(
                    || UserError(format!("Parts you can cut: {}",
                                         &join_words(&parts_remaining.iter().filter_map(|pt| possession_data().get(pt))
                                                    .map(|pd| pd.display).collect::<Vec<&'static str>>())
                    )))?
        };

        let possession_data = possession_data().get(possession_type)
            .ok_or_else(|| UserError("That part doesn't exist anymore".to_owned()))?;
        
        let mut corpse_mut = (*corpse).clone();
        match corpse_mut.death_data.as_mut() {
            None => {},
            Some(dd) => {
                dd.parts_remaining = 
                    dd
                    .parts_remaining
                    .iter().take_while(|pt| pt != &possession_type)
                    .chain(dd.parts_remaining.iter().skip_while(|pt| pt != &possession_type).skip(1))
                    .map(|pt| (*pt).clone())
                    .collect()
            }
        }

        match check_item_capacity(&ctx.trans, &player_item.refstr(), possession_data.weight).await? {
            CapacityLevel::AboveItemLimit | CapacityLevel::OverBurdened =>
                user_error("You have too much stuff to take that on!".to_owned())?,
            _ => {}
        }
        
        if corpse_mut.death_data.as_ref().map(|dd| dd.parts_remaining.is_empty()) == Some(true) {
            destroy_container(&ctx.trans, &corpse_mut).await?;
        } else {
            ctx.trans.save_item_model(&corpse_mut).await?;
        }

        let mut player_item_mut = (*player_item).clone();
        if skill_check_and_grind(&ctx.trans, &mut player_item_mut, &SkillType::Craft, 10.0).await? < 0.0 {
            broadcast_to_room(&ctx.trans, &player_item.location, None,
                              &format!("{} tries to cut the {} from {}, but only leaves a mutilated mess.\n",
                                       &player_item.display_for_sentence(true, 1, true),
                                       possession_data.display,
                                       corpse.display_for_sentence(true, 1, false)
                              ),
                              Some(&format!("{} tries to cut the {} from {}, but only leaves a mutilated mess.\n",
                                            &player_item.display_for_sentence(true, 1, true),
                                            possession_data.display,
                                            corpse.display_for_sentence(true, 1, false)
                              ))
            ).await?;
        } else {
            let mut new_item: Item = (*possession_type).clone().into();
            new_item.item_code = format!("{}", ctx.trans.alloc_item_code().await?);
            new_item.location = player_item.refstr();
            ctx.trans.save_item_model(&new_item).await?;

            broadcast_to_room(&ctx.trans, &player_item.location, None,
                              &format!("{} expertly cuts the {} from {}.\n",
                                       &player_item.display_for_sentence(true, 1, true),
                                       possession_data.display,
                                       corpse.display_for_sentence(true, 1, false)
                              ),
                              Some(&format!("{} expertly cuts the {} from {}.\n",
                                            &player_item.display_for_sentence(true, 1, true),
                                            possession_data.display,
                                            corpse.display_for_sentence(true, 1, false)
                              ))
            ).await?;
        }

        ctx.trans.save_item_model(&player_item_mut).await?;

        Ok(())
    }
}
static VERB_INT: Verb = Verb;
pub static VERB: UserVerbRef = &VERB_INT as UserVerbRef;
